{{
'-------------------------------------------------------------------------------
'
'
' Catalina_HUB_XMM_Loader - HUB Loader for use in conjunction with the Catalina
'                           Catalina code generator backend to LCC
'
' Note this loader assumes the Catalina image is already in XMM.
' It leaves the code segment in situ, but moves other segments
' to Hub RAM (x2-x4). Then it loads and starts the kernel. 
'
' Version 1.0 - Initial version by Ross Higson.
' Version 1.1 - Support for large address model (using layout -x5)
' Version 1.2 - Add Morpheus support.
'             - Added Run as an alternative way of starting the loader
' Version 2.5 - Add RamBlade support and minor tweaks to XMM code. 
'
' This program incorporates XMM software derived from:
'
'    Cluso's TriBladeProp Blade #2 Driver v0.201
'
'    HX512SRAM_ASM_DRV_101.spin by Andre' LaMothe, 
'       (C) 2007 Nurve Networks LLC
'                              
'    Morpheus System Architecture and Developer's Guide v0.90
'       (C) 2009 William Henning
'
'-------------------------------------------------------------------------------
}}
CON
'
' Common Constants
'
PAGE_SIZE   = 256  ' Size of page in bytes. Fairly arbitrary,
                   ' but must be at least 64 and divisible by 4.
                  
INIT_BZ_OFF = Common#LMM_INIT_BZ_OFF
INIT_PC_OFF = Common#LMM_INIT_PC_OFF

MAX_KERNEL_LONGS = $1f0  ' number of kernel longs to copy

XMM_BASE_ADDRESS = common#XMM_BASE_ADDRESS

CON 

' HX512 XMM Constants (for Hybrid or Hydra only):

HX_DBUS_BIT_SHIFT    = 16
HX_CTRL_BIT_SHIFT    = 1
HX_STROBE_BIT_SHIFT  = 30

' commands for controller, pre-shifted to left 1-bit, so we don't have to do it
' during runtime, small enough to fit in constants
HX_CMD_WRITE_SHIFT   = %00000_00_0
HX_CMD_READ_SHIFT    = %00000_01_0
HX_CMD_LOADLO_SHIFT  = %00000_10_0
HX_CMD_LOADHI_SHIFT  = %00000_11_0

' sram commands
HX_CMD_WRITE   = %000000_00
HX_CMD_READ    = %000000_01
HX_CMD_LOADLO  = %000000_10
HX_CMD_LOADHI  = %000000_11

OBJ
  Common   : "Catalina_Common"                          ' Common Definitions
  
PUB Start(Registry_Addr, Stack_Top, Kernel_Addr, Page_Buffer) : Cog | Reg, Stack, Kernel, Page

   ' Start the loader in another cog, and return the cog that was used, or -1 if no cog is available.

   Reg    := Registry_Addr
   Stack  := Stack_Top
   Kernel := Kernel_Addr
   Page   := Page_Buffer

   Cog := cognew(@entry, @Reg)

PUB Run(Registry_Addr, Stack_Top, Kernel_Addr, Page_Buffer) : Cog | Reg, Stack, Kernel, Page

   ' Run restarts this cog as the loader, and never returns

   Reg    := Registry_Addr
   Stack  := Stack_Top
   Kernel := Kernel_Addr
   Page   := Page_Buffer

   coginit(cogid, @entry, @Reg)

DAT
        org 0

entry
        mov    r0,par                           ' point to parameters
        rdlong reg_addr,r0                      ' get address of Registry
        add    r0,#4
        rdlong stack_addr,r0                    ' get address of top of stack
        add    r0,#4
        rdlong entry_addr,r0                    ' get kernel entry address
        add    r0,#4
        rdlong page_addr,r0                     ' get address of page buffer (long aligned)

        call   #XMM_Activate

        ' retrieve the initial BZ, PC, layout and segment addresses 
        ' from the image
        '
        ' Note: the initial BZ and PC in the image will be correct    
        ' once the data and code segments are moved - the BZ value
        ' is correct if we load the data segment to the place the
        ' compiler assumed it would be loaded, and the XMM Kernel
        ' adjusts the PC value to subtract the location of the code
        ' segment, assuming it will end up at location zero in XMM.
        ' If this is not the case, the intial PC value will need to
        ' be adjusted. Note all internal code segment references are
        ' relocated "on the fly" by the XMM Kernel.
        '

        call   #ClearPage                       ' for debug only - make sure page buffer is empty
        mov    XMM_Addr,#$10 + INIT_BZ_OFF<<2 + 8 ' point to initial BZ (offset $144)
        mov    Hub_Addr,page_addr
        mov    XMM_Len,#PAGE_SIZE
        call   #XMM_ReadPage
        mov    r0,page_addr

        rdlong Init_BZ,r0
        add    r0,#4                             
        rdlong Init_PC,r0
        add    r0,#4                             
        rdlong layout,r0
        add    r0,#4
        rdlong code_addr,r0
        add    r0,#4
        rdlong cnst_addr,r0
        add    r0,#4
        rdlong init_addr,r0
        add    r0,#4
        rdlong data_addr,r0
        add    r0,#4
        rdlong ends_addr,r0

'
' NOTE: the following code implements:
' XMM Layout 2:
'    [CNST,INIT,DATA,CODE] => [CODE] in XMM & [CNST,INIT,DATA] in HUB
' XMM Layout 3:
'    [INIT,DATA,CODE,CNST] => [CODE,CNST] in XMM & [INIT,DATA] in HUB
' XMM Layout 4:
'    [DATA,CODE,CNST,INIT] => [CODE,CNST,INIT] in XMM & [DATA] in HUB
' XMM Layout 5:
'    [CODE,CNST,INIT,DATA] => [CODE,CNST,INIT,DATA] in XMM (stack in HUB)
'
        ' Copy MAX_KERNEL_LONGS*4 bytes from the Kernel entry address  
        ' to ends_addr in XMM RAM - this is just a temporary place to
        ' save the kernel so we don't lose it when we overwrite the Hub
        ' RAM from the XMM RAM. Note that for layout 5 we take into 
        ' account the bytes removed from the image we have in XMM
         
        mov    byte_count,kernel_max
        mov    src_addr,entry_addr
        mov    dst_addr,ends_addr
        add    dst_addr,#$10
        cmp    layout,#5 wz                     ' for layout 5, correct  ...
  if_z  sub    dst_addr,l5size                  ' ... for bytes removed

        call   #Copy_RAM_To_XMM

        ' Then copy the first $200 bytes of image from XMM to RAM. Note that
        ' we don't overwrite the first $10 bytes since they're special - i.e.
        ' clock freq etc - so we only copy from byte $10 onwards

        mov    byte_count,l5offs   
        sub    byte_count,#$10
        mov    src_addr,#$10 
        mov    dst_addr,#$10

        call   #Copy_XMM_To_RAM

        ' Now relocate the segments from XMM to RAM (depending on the layout)

        cmp    layout,#5 wz                     ' for layout 5 ...
  if_z  jmp    #reloc_done                      ' ... no segments to relocate
        mov    byte_count,Init_BZ               ' for other layouts we copy (up to) Init_BZ
        cmp    layout,#4 wz                     ' for layout 4 ...
  if_z  mov    src_addr,data_addr               ' ... start from data segment
        cmp    layout,#3 wz                     ' for layout 3 ...            
  if_z  mov    src_addr,init_addr               ' ... start from init segment 
        cmp    layout,#2 wz                     ' for layout 2 ...            
  if_z  mov    src_addr,cnst_addr               ' ... start from cnst segment 

        sub    byte_count,src_addr              ' calculate size to copy
        mov    dst_addr,src_addr                ' destination addr in RAM is correct ...
        add    src_addr,#$10                    ' ... but we must correct source addr

        call   #Copy_XMM_To_RAM
        
reloc_done

        ' Now we must move the code segment to start at XMM location
        ' zero as the XMM Kernel expects. 

        mov    src_addr,code_addr               ' code segment at code_addr ...
        add    src_addr,#$10                    ' ... corrected ...
        mov    dst_addr,#$0                     ' ... goes at offset 0 in XMM
        cmp    layout,#5 wz                     ' for layout 5, correct ...
  if_z  sub    src_addr,l5size                  ' ... for bytes removed ...
        mov    byte_count,ends_addr
        sub    byte_count,code_addr

        call   #Copy_XMM_To_XMM

{
        call  #XMM_Check
}        

        ' Copy MAX_KERNEL_LONGS*4 bytes from ends_addr to straight after
        ' Init_BZ (layouts 2,3,4) or after the segment table (layout 5) - this 
        ' is just another temporary place to put the kernel code - later,
        ' we will overwrite ourselves with this code when we restart this
        ' cog.
        
        mov    byte_count,kernel_max
        mov    src_addr,ends_addr
        add    src_addr,#$10
        cmp    layout,#5 wz                     ' for layout 5 we correct ...
  if_z  sub    src_addr,l5size                  ' ... for bytes removed ...
  if_z  mov    dst_addr,#$10+INIT_PC_OFF<<2+8+28 ' ... and copy after seg table
  if_nz mov    dst_addr,Init_BZ                 ' for other layouts ...
  if_nz add    dst_addr,#$10                    ' ... copy to Init_BZ (+$0010)

        call   #Copy_XMM_To_RAM

        ' Set up the Catalina program base address for use by the Kernel.

        cogid   r0                              ' convert ...
        shl     r0,#2                           ' ... my cog id ...
        add     r0,reg_addr                     ' ... to my registration addr
        rdlong  r2,r0                           ' get my request block addr

        add     r2,#4
        wrlong  stack_addr,r2
        sub     r2,#4
        mov     r1,#$10                         ' initial BA is always $10
        wrlong  r1,r2

        ' Then become the Kernel by restarting the cog with the Kernel code.

        cogid   r0                              ' set the cog id
        cmp     layout,#5 wz                    ' for layout 5 ...
   if_z mov     r1,#$10+INIT_PC_OFF<<2+8+28     ' ... kernel is after seg table
  if_nz mov     r1,Init_BZ                      ' for other layouts ...
  if_nz add     r1,#$10                         ' ... it is at Init_BZ (+$0010)
        shl     r1,#2
        or      r0,r1                           ' set the code address
        mov     r1,reg_addr
        shl     r1,#16
        or      r0,r1                           ' set the par address

       ' call XMM_TriState to avoid side effects (e.g. resetting 
       ' other CPUs on the TriBladeProp) while restarting

        call #XMM_Tristate

        coginit r0

        ' the following constants are used when reconstructing layout 5:

l5size  long    $8000 - $0200                   ' this many bytes removed
l5offs  long    $0200                           ' offset of code segment

low24   long    $00FFFFFF

'-------------------------------- Utility routines -----------------------------
'
Copy_XMM_To_RAM
        tjz    byte_count,#Copy_XMM_To_RAM_ret
        mov    XMM_Addr,src_addr
        mov    Hub_Addr,dst_addr
        mov    XMM_Len,byte_count
        call   #XMM_ReadPage
Copy_XMM_To_RAM_ret
        ret

Copy_RAM_To_XMM
        tjz    byte_count,#Copy_RAM_To_XMM_ret
        mov    XMM_Addr,dst_addr
        mov    Hub_Addr,src_addr
        mov    XMM_Len,byte_count
        call   #XMM_WritePage
Copy_RAM_To_XMM_ret
        ret

Copy_RAM_To_RAM
        tjz    byte_count,#Copy_RAM_To_RAM_ret
:Copy_loop
        rdbyte r0,src_addr
        wrbyte r0,dst_addr
        add    src_addr,#1
        add    dst_addr,#1
        djnz   byte_count,#:Copy_loop
Copy_RAM_To_RAM_ret
        ret

Copy_XMM_To_XMM
        tjz    byte_count,#Copy_XMM_To_XMM_ret
        movd   XMM_Dst,#XMM_XMM_Tmp
        movs   XMM_Src,#XMM_XMM_Tmp
:Copy_loop
        mov    XMM_Addr,src_addr
        call   #XMM_ReadLong
        mov    XMM_Addr,dst_addr
        call   #XMM_WriteLong
        add    src_addr,#4
        add    dst_addr,#4
        sub    byte_count,#4 wz
 if_nz  jmp    #:Copy_loop
Copy_XMM_To_XMM_ret
        ret
XMM_XMM_Tmp long 0
'
'f_d32u - Unsigned 32 bit division
'         Divisor : r1
'         Dividend : r0
'         Result:
'             Quotient in r0
'             Remainder in r1

f_d32u
        mov    ftemp,#32
        mov    ftmp2, #0
:up2
        shl    r0,#1       WC
        rcl    ftmp2,#1    WC
        cmp    r1,ftmp2    WC,WZ
 if_a   jmp    #:down
        sub    ftmp2,r1
        add    r0,#1
:down
        sub    ftemp, #1   WZ
 if_ne  jmp    #:up2
        mov    r1,ftmp2
f_d32u_ret
        ret


'
'f_m32 - multiplication
'        r0 : 1st operand (32 bit)
'        r1 : 2nd operand (32 bit)
'        Result:
'           Product in r0 (<= 32 bit)
'
f_m32
        mov    ftemp,#0
:start
        cmp    r0,#0       WZ
 if_e   jmp    #:down3
        shr    r0,#1       WC
 if_ae  jmp    #:down2
        add    ftemp,r1    WC
:down2
        shl    r1,#1       WC
        jmp    #:start
:down3
        mov    r0,ftemp
f_m32_ret
        ret

ClearPage  
        mov    r0,#0
        mov    r1,#PAGE_SIZE
        mov    r2,page_addr
:Clear_loop
        wrbyte r0,r2
        add    r2,#1
        djnz   r1,#:Clear_loop
ClearPage_ret
        ret


'-------------------------------- DEBUG routines -------------------------------
{
XMM_Check
        mov   XMM_Addr,CheckAddr
        movd  XMM_Dst,#CheckRslt
        call  #XMM_ReadLong
        mov   hex_value,CheckRslt
        call  #Flash_Hex
XMM_Check_ret
        ret        
        

CheckAddr     long      $4ba18 - ($8000 - $200)         ' correct for bytes removed
CheckRslt     long      $0        
}
{
'
#include "debug.inc"
'
}
' Common variables

r0            long      $0
r1            long      $0
r2            long      $0
r3            long      $0
r4            long      $0
r5            long      $0
r6            long      $0

page_addr     long      $0
reg_addr      long      $0
stack_addr    long      $0
entry_addr    long      $0

Init_BZ       long      $0
Init_PC       long      $0

code_addr     long      $0
cnst_addr     long      $0
init_addr     long      $0
data_addr     long      $0
ends_addr     long      $0

layout        long      $0
code_size     long      $0
cnst_size     long      $0
init_size     long      $0
data_size     long      $0

src_addr      long      $0
dst_addr      long      $0
page_count    long      $0
byte_count    long      $0
'
' temporary storage used in mul & div calculations
'
ftemp         long      $0
ftmp2         long      $0
ftmp3         long      $0

kernel_max    long      MAX_KERNEL_LONGS*4
'
'=============================== XMM SUPPORT CODE ==============================
'
' The folling defines determine which XMM functions are included - comment out
' the appropriate lines to exclude the corresponding XMM function:
'
#define NEED_XMM_READLONG
#define NEED_XMM_WRITELONG
#define NEED_XMM_READPAGE
#define NEED_XMM_WRITEPAGE
#define NEED_XMM_INCADDR        ' required for XMM_ReadPage and XMM_WritePage
#define ACTIVATE_INITS_XMM      ' the HX512 does not allow this in the Kernel
'                                    
#include "XMM.inc"
'
'============================ END OF XMM SUPPORT CODE ==========================
'
              fit       $1f0                    ' max size
'              
{{
                            TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}