{{
'-------------------------------------------------------------------------------
'
'   Catalina_SD_Plugin - an SD Card plugin for Catalina.
'
'   Modified to be a Catalina plugin (removal of Vars and replacement
'   with registry, add ability to use non-consecutive pins) by
'   Ross Higson, based on:
'
'   sdspi:  SPI interface to a Secure Digital card.
'
'   WARNING: Due to the need to coordinate the enabling and disabling of the
'   SD Card and the XMM RAM on platforms that share pins, this driver can be
'   very fiddly to modify.  
'
'   Original notes by Radical Eye Software are given below ...
'
'-------------------------------------------------------------------------------
'
'   Copyright 2008   Radical Eye Software
'
'   See end of file for terms of use.
'
'   You probably never want to call this; you want to use fsrw
'   instead (which calls this); this is only the lowest layer.
'
'   Assumes SD card is interfaced using four consecutive Propeller
'   pins, as follows (assuming the base pin is pin 0):
'                3.3v
'              ? ? ? ? ? ?
'                    20k
'   p0 ????????????????????????? do
'   p1 ????????????????????????? clk
'   p2 ????????????????????????? di
'   p3 ????????????????????????? cs (dat3)
'         150          ????????? irq (dat1)
'                        ??????? p9 (dat2)
'
'   The 20k resistors
'   are pullups, and should be there on all six lines (even
'   the ones we don't drive).
'
'   This code is not general-purpose SPI code; it's very specific
'   to reading SD cards, although it can be used as an example.
'
'   The code does not use CRC at the moment (this is the default).
'   With some additional effort we can probe the card to see if it
'   supports CRC, and if so, turn it on.   
'
'   All operations are guarded by a watchdog timer, just in case
'   no card is plugged in or something else is wrong.  If an
'   operation does not complete in one second it is aborted.
'
'-------------------------------------------------------------------------------
'
' Version 1.0 - Initial version by Ross Higson.
'
' Version 2.9 - added C3 support, and tidied up shared platform support (i.e.
'               TRIBLADEPROP, RAMBLADE & c3) as much as possible.
}}
'
' define SHARED_SD on platforms that must share the SD Card pins with other devices:
'
#ifdef TRIBLADEPROP
#define SHARED_SD
#elseifdef RAMBLADE
#define SHARED_SD
#elseifdef C3
#define SHARED_SD
#endif
'
'#define DEBUG
'#define DEBUG_REQUESTS
'#define DEBUG_READS
'#define DEBUG_CMD_RESPONSE

con

sectorsize  = 512
sectorshift = 9

SD_Init   = 1
SD_Read   = 2
SD_Write  = 3
SD_ByteIO = 4
SD_StopIO = 5

' SD services:
'
' The command to perform is encoded in the top 8 bits of the parameter
' The address of a parameter block is encoded in the bottom 24 bits.
' The parameter block is 2 longs:
'    - the buffer adress to use
'    - the sector number to read/write 

'name: SD_Init - Initialize the driver
'code: 1
'type: long request
'data: parameter block address
'rslt: (none)
'
'name: SD_Read - Read a sector
'code: 2
'type: long request
'data: parameter block address
'rslt: (none)
'
'name: SD_Write - Write a sector
'code: 3 
'type: long request
'data: parameter block address
'rslt: (none)
   
'name: SD_ByteIO - Write a sector
'code: 4 
'type: short request
'data: byte to write
'rslt: (none)

'name: SD_StopIO - disable the SD card (required on the TRIBLADEPROP and RAMBLADE)
'code: 5 
'type: short request
'data: byte to write
'rslt: (none)
   
LMM_FIL     = common#LMM_FIL

obj
   common : "Catalina_Common"
   
pub start(registry)  : ok | sd_cog
'
'   Initialize the card!  Send a whole bunch of
'   clocks (in case the previous program crashed
'   in the middle of a read command or something),
'   then a reset command, and then wait until the
'   card goes idle.
'
  do  := Common#SD_DO_PIN
  clk := Common#SD_CLK_PIN 
  di  := Common#SD_DI_PIN
#ifdef C3
  cs_clr := Common#SD_CS_CLR
  cs_clk := Common#SD_CS_CLK
#else
  cs  := Common#SD_CS_PIN
#endif
  clockfreq := Common#CLOCKFREQ

  sd_cog := cognew(@entry, registry)
  
  if sd_cog > 0
  
    ' wait till the sd plugin has registered
    repeat until long[registry][sd_cog] & $FF000000 <> 0

    ' wait for intialization to complete       
    common.WaitForRequest(sd_cog)

    ok := sd_cog + 1
    
  else
  
    ok := 0

    
dat

        org
entry
        cogid   acca                            ' get ...
        shl     acca,#2                         ' ... our ...
        add     acca,par                        ' ... registry block entry
        rdlong  rqstptr,acca                    ' register ...
        and     rqstptr,Mask24                  ' ... this ...
        mov     accb,#LMM_FIL                   ' ... plugin ...
        shl     accb,#24                        ' ... as the ...
        or      accb,rqstptr                    ' ... appropriate ...
        wrlong  accb,acca                       ' ... type
        mov     rsltptr,rqstptr                 ' set up a pointer to ...
        add     rsltptr,#4                      ' ... our result address
        neg     acca,#1                         ' indicate ...
        wrlong  acca,rqstptr                    ' ... initialization in progress

        mov     domask,#1
        shl     domask,do
        mov     dimask,#1
        shl     dimask,di
        mov     clkmask,#1
        shl     clkmask,clk
#ifdef C3          
        mov     cs_clrmsk,#1
        shl     cs_clrmsk,cs_clr
        mov     cs_clkmsk,#1
        shl     cs_clkmsk,cs_clk
#else          
        mov     csmask,#1
        shl     csmask,cs
#endif

initialize

        call    #SD_activate

        neg     phsb,#1
        mov     frqb,#0
        mov     acca,nco
        add     acca,clk
        mov     ctra,acca
        mov     acca,nco
        add     acca,di
        mov     ctrb,acca
        mov     ctr2,onek
oneloop        
        call    #sendiohi
        djnz    ctr2,#oneloop
        mov     starttime,cnt
        mov     cmdo,#0
        mov     cmdp,#0
        call    #cmd
        
#ifdef RAMBLADE
        or      outa,csmask
#else
        call    #SD_tristate
#endif

        call    #sendiohi
initloop       

        call    #SD_activate

        mov     cmdo,#55
        call    #cmd
        mov     cmdo,#41
        call    #cmd
        
#ifdef RAMBLADE
        or      outa,csmask
#else
        call    #SD_tristate
#endif

        cmp     accb,#1 wz
   if_z jmp     #initloop
        wrlong  accb,rsltptr

' reset frqa and the clock
finished
        mov     frqa,#0
#ifdef TRIBLADEPROP
        or      outa,csmask
        neg     phsb,#1
        call    #sendiohi
        call    #SD_tristate
        mov     acca,#511                       ' on the TriBlade, we must not let the 
        add     acca,cnt                        ' kernel resume until we have paused
        waitcnt acca,#0                         ' for a short period after disabling the
        mov     acca,#0                         ' SD card, or the XMM access goes haywire. 
        wrlong  acca,rqstptr                    ' (this should do no harm on other cards).                     
        jmp     #waitloop                           
#elseifdef RAMBLADE
        or      outa,csmask
        neg     phsb,#1
        call    #sendiohi
        call    #SD_tristate
        mov     acca,#511                       ' on the RamBlade, we must not let the 
        add     acca,cnt                        ' kernel resume until we have paused
        waitcnt acca,#0                         ' for a short period after disabling the
        mov     acca,#0                         ' SD card, or the XMM access goes haywire. 
        wrlong  acca,rqstptr                    ' (this should do no harm on other cards).                     
        jmp     #waitloop                           
#elseifdef C3
        call    #SD_tristate
        neg     phsb,#1
        call    #sendiohi
        mov     acca,#511                       ' on the C3, we must not let the 
        add     acca,cnt                        ' kernel resume until we have paused
        waitcnt acca,#0                         ' for a short period after disabling the
        mov     acca,#0                         ' SD card, or the XMM access goes haywire. 
        wrlong  acca,rqstptr                    ' (this should do no harm on other cards).                     
        jmp     #waitloop                           
#else
        wrlong  frqa,rqstptr
        or      outa,csmask
        neg     phsb,#1
        call    #sendiohi
#endif

pause
        mov     acca,#511
        add     acca,cnt
        waitcnt acca,#0

waitloop
        mov     starttime,cnt
        rdlong  acca,rqstptr wz
   if_z jmp     #pause
        mov     paramptr,acca
        and     paramptr,Mask24
        rdlong  buffptr,paramptr
        add     paramptr,#4
        shr     acca,#24
#ifdef DEBUG_REQUESTS
        mov     hex_value,acca
        call    #Flash_Hex
#endif   
        cmp     acca,#SD_Init wz
   if_z jmp     #initialize
        cmp     acca,#SD_ByteIO wz
   if_z jmp     #byteio
        mov     ctr2,sector
        cmp     acca,#SD_Read wz
   if_z jmp     #rblock
        cmp     acca,#SD_Write wz
   if_z jmp     #wblock
        cmp     acca,#SD_StopIO wz
   if_z jmp     #finished
        jmp     #pause
  
wblock
#ifdef SHARED_SD
        call    #SD_activate
#endif
        mov     starttime,cnt
        mov     cmdo,#24
        rdlong  cmdp,paramptr
        call    #cmd
        mov     phsb,#$fe
        call    #sendio
        mov     accb,buffptr
        neg     frqa,#1
wbyte
        rdbyte  phsb,accb
        shl     phsb,#23
        add     accb,#1
        mov     ctr,#8
wbit    mov     phsa,#8
        shl     phsb,#1
        djnz    ctr,#wbit
        djnz    ctr2,#wbyte        
        neg     phsb,#1
        call    #sendiohi
        call    #sendiohi
        call    #readresp
        and     accb,#$1f
        sub     accb,#5
        wrlong  accb,rsltptr
        call    #busy
        jmp     #finished

rblock
#ifdef SHARED_SD
        call    #SD_activate
#endif
        mov     starttime,cnt
        mov     cmdo,#17
        rdlong  cmdp,paramptr
        call    #cmd
        call    #readresp
        mov     accb,buffptr
        sub     accb,#1
rbyte
        mov     phsa,hifreq
        mov     frqa,freq
        add     accb,#1
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        test    domask,ina wc
        addx    acca,acca
        mov     frqa,#0
        test    domask,ina wc
        addx    acca,acca
        wrbyte  acca,accb
        djnz    ctr2,#rbyte        
        mov     frqa,#0
        neg     phsb,#1
        call    #sendiohi
        call    #sendiohi

#ifdef DEBUG_READS
        rdbyte hex_value,buffptr
        call #Flash_Hex
        add buffptr,#1        
        rdbyte hex_value,buffptr
        call #Flash_Hex
        add buffptr,#1        
        rdbyte hex_value,buffptr
        call #Flash_Hex
        add buffptr,#1        
        rdbyte hex_value,buffptr
        call #Flash_Hex
#endif

#ifndef C3
        or      outa,csmask
#endif        
        wrlong  ctr2,rsltptr
        jmp     #finished

byteio     
#ifdef SHARED_SD
        call    #SD_activate
#endif
        mov     phsb,buffptr
        call    #sendio
        wrlong  accb,rsltptr
        jmp     #finished

sendio
        rol     phsb,#24
sendiohi
        mov     ctr,#8
        neg     frqa,#1
        mov     accb,#0
bit     mov     phsa,#8
        test    domask,ina wc
        addx    accb,accb        
        rol     phsb,#1
        djnz    ctr,#bit
sendio_ret
sendiohi_ret
        ret

checktime
        mov     duration,cnt
        sub     duration,starttime
        cmp     duration,clockfreq wc
checktime_ret
  if_c  ret
        neg     duration,#13
        wrlong  duration,rsltptr
        jmp     #finished

cmd
#ifdef DEBUG_CMD_RESPONSE
        mov hex_value,cmdo
        and hex_value,#$FF
        call #Flash_Hex
#endif        
#ifndef C3
        andn    outa,csmask
#endif        
        neg     phsb,#1
        call    #sendiohi
        mov     phsb,cmdo
        add     phsb,#$40
        call    #sendio
        mov     phsb,cmdp
        shl     phsb,#9
        call    #sendiohi
        call    #sendiohi
        call    #sendiohi
        call    #sendiohi
        mov     phsb,#$95
        call    #sendio
readresp
        neg     phsb,#1
        call    #sendiohi
#ifndef DEBUG        
        call    #checktime
#endif        
        cmp     accb,#$ff wz
   if_z jmp     #readresp 
#ifdef DEBUG_CMD_RESPONSE
        mov hex_value,accb
        and hex_value,#$FF
        call #Flash_Hex
#endif        
cmd_ret
readresp_ret
        ret

busy
        neg     phsb,#1
        call    #sendiohi
        call    #checktime
        cmp     accb,#$0 wz
   if_z jmp     #busy
busy_ret
        ret
'
' SD_activate - set up control lines for SD Card and select it
'
SD_activate 
{
        call  #LED_On
}
#ifdef TRIBLADEPROP
        mov   acca,ram_disable
        or    acca,ram_LE_bit
        mov   outa,acca
        mov   dira,acca
        or    dira,dimask
        or    dira,clkmask
        or    dira,csmask
        andn  outa,csmask
#elseifdef C3
        or    dira,dimask
        or    dira,clkmask
        andn  outa,cs_clkmsk
        or    outa,cs_clrmsk
        or    dira,cs_clkmsk
        or    dira,cs_clrmsk
        andn  outa,cs_clrmsk
        or    outa,cs_clrmsk
        andn  outa,cs_clkmsk
        mov   cs_cnt,#5
:selectSPI                        
        or    outa,cs_clkmsk
        andn  outa,cs_clkmsk
        djnz  cs_cnt,#:selectSPI
#else
        or    dira,dimask
        or    dira,clkmask
        andn  outa,csmask
        or    dira,csmask
#endif
SD_activate_ret
        ret   
'
'SD_Tristate : Give up the control lines shared between SD and XMM, delect SD
'
SD_tristate 
{
        call    #LED_Off
}        
#ifdef TRIBLADEPROP
        mov     outa,#0
        mov     dira,#0
#elseifdef RAMBLADE
        mov     outa,#0
        mov     dira,#0
#elseifdef C3
        andn    outa,cs_clrmsk
        or      outa,cs_clrmsk
        mov     dira,#0
#else        
        or      outa,csmask
#endif        
SD_tristate_ret
        ret
'
#ifdef TRIBLADEPROP
'
' TriBladeProp Blade #2 Hardware Access:
'
'                                        +-------- -CS (U26 FLASH)   -+                                    
'                                        |+------- -CS (J22 microSD)  +- Latched pins...                   
'                                        ||+------ -RST3              |    passes when LE=1                
'           (P30) SO ---+                |||+----- -RST1              |    latches on  LE low going edge   
'           (P31) SI --+|                ||||+---- -CE (U24 SRAM-hi)  |    latched on  LE=0                
'                      ||+---- -WE(SDA)  |||||+--- -CE (U23 SRAM-lo)  |                                    
'                      |||+--- -OE(SCL)  ||||||+-- A20                |                                    
'                      ||||+--  LE       |||||||+- A19               -+                                    
'                      |||||             ||||||||                                                          
ram_LE_bit    long    %00001_00000000000_00000000_00000000 ' LE bit
ram_disable   long    %00000_00000000000_00111100_00000000 ' -RST1=1, -RST3=1, -CE(U24)=1, -CE(U23)=1                                 
'                            +---------+ +------+ +------+
'                            A18......A8 A7....A0 D7....D0
'                                        Q7....Q0
#endif

#ifdef DEBUG
'
' include some useful LED debugging routines
'
#include "debug.inc"
'
#endif
'
di        long 0
do        long 0
clk       long 0
#ifdef C3
cs_clr    long 0
cs_clk    long 0
cs_cnt    long 0
#else
cs        long 0
#endif

clockfreq long 0

rqstptr   long 0
rsltptr   long 0
paramptr  long 0
buffptr   long 0
param2    long 0


nco       long $1000_0000
hifreq    long $e0_00_00_00
freq      long $20_00_00_00

onek      long 1000
sector    long 512

Zero      long 0
Mask24    long $00FF_FFFF

domask    res 1
dimask    res 1
clkmask   res 1
#ifdef C3
cs_clrmsk res 1
cs_clkmsk res 1
#else
csmask    res 1
#endif
acca      res 1
accb      res 1
cmdo      res 1
cmdp      res 1
ctr       res 1
ctr2      res 1
starttime res 1
duration  res 1
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}