{{
'-------------------------------------------------------------------------------
'
' XMM_EEPROM_Loader - XMM EEPROM Loader for use in conjunction
'                     with the Catalina code generator
'
' Note that this version supports programs built using -x2 to -x5. The 
' image is assumed to be located in eeprom at the location given on
' startup (usually $8000) 
'
' Version 1.0 - Initial version by Ross Higson.
' Version 1.1 - Load kernel from target, not program.
'               Support for x2, x3, x4 and x5 layouts.
' Version 1.3 - Support for large address model (using layout -x5)
' Version 1.4 - Add Morpheus support.
'             - Added Run as an alternative way of starting the loader
' Version 2.5 - Add RamBlade support and minor tweaks to XMM code. 
'
' Incorporates I2C code based on:
'
'    Basic_I2C_Driver.spin by Parallax and Michael Green
'
' This program incorporates XMM software derived from:
'
'    Cluso's TriBladeProp Blade #2 Driver v0.201
'
'    HX512SRAM_ASM_DRV_101.spin by Andre' LaMothe, 
'       (C) 2007 Nurve Networks LLC
'
'    Morpheus System Architecture and Developer's Guide v0.90
'       (C) 2009 William Henning
'
'---------------------------------------------------------------
}}
CON
'
' Common Constants
'
PAGE_SIZE   = 32  ' Size of page in bytes. Must be divisible by 4.
                  ' Note that a page size > 32 may not work with
                  ' some types of eeprom
                  
INIT_BZ_OFF = Common#LMM_INIT_BZ_OFF
INIT_PC_OFF = Common#LMM_INIT_PC_OFF

MAX_KERNEL_LONGS = $1f0  ' number of kernel longs to copy

XMM_BASE_ADDRESS = common#XMM_BASE_ADDRESS

CON
'
' I2C (EEPROM) Constants:
'
ACK      = 0                        ' I2C Acknowledge
NAK      = 1                        ' I2C No Acknowledge
Xmit     = 0                        ' I2C Direction Transmit
Recv     = 1                        ' I2C Direction Receive

I2C_PIN  = Common#I2C_PIN           ' I2C Boot EEPROM SCL Pin

I2C_DEV  = Common#I2C_DEV           ' I2C Boot EEPROM Device Address

I2C_DELAY_COUNT = 0                 ' clock counts to delay per I2C bit (on the
                                    ' Hydra this driver works with minimal delay,
                                    ' but higher values may be needed on other
                                    ' systems)

CON

' HX512 XMM Constants (for Hybrid or Hydra only):

HX_DBUS_BIT_SHIFT    = 16
HX_CTRL_BIT_SHIFT    = 1
HX_STROBE_BIT_SHIFT  = 30

' commands for controller, pre-shifted to left 1-bit, so we don't have to do it
' during runtime, small enough to fit in constants
HX_CMD_WRITE_SHIFT   = %00000_00_0
HX_CMD_READ_SHIFT    = %00000_01_0
HX_CMD_LOADLO_SHIFT  = %00000_10_0
HX_CMD_LOADHI_SHIFT  = %00000_11_0

' sram commands
HX_CMD_WRITE   = %000000_00
HX_CMD_READ    = %000000_01
HX_CMD_LOADLO  = %000000_10
HX_CMD_LOADHI  = %000000_11


OBJ
  Common   : "Catalina_Common"                          ' Common Definitions
  
PUB Start(Registry_Addr, Stack_Top, Kernel_Addr, Image_Addr, Page_Buffer) : Cog | Pin, Device, Registry, Stack, Kernel, Image, Page

   ' Start the loader in another cog, and return the cog that was used, or -1 if no cog is available.

   Pin      := I2C_PIN
   Device   := I2C_DEV 
   Registry := Registry_Addr
   Stack    := Stack_Top
   Kernel   := Kernel_Addr
   Image    := Image_Addr
   Page     := Page_Buffer

   Cog := cognew(@entry, @Pin)

PUB Run(Registry_Addr, Stack_Top, Kernel_Addr, Image_Addr, Page_Buffer) : Cog | Pin, Device, Registry, Stack, Kernel, Image, Page

   ' Run restarts this cog as the loader, and never returns

   Pin      := I2C_PIN
   Device   := I2C_DEV 
   Registry := Registry_Addr
   Stack    := Stack_Top
   Kernel   := Kernel_Addr
   Image    := Image_Addr
   Page     := Page_Buffer

   coginit(cogid, @entry, @Pin)

DAT
        org 0

entry
        mov    r0,par                           ' point to parameters
        rdlong SCL,r0                           ' get I2C Clock Pin (Data Pin is assumed to be Clock Pin + 1)
        mov    SDA,SCL                          ' set up pins and masks
        add    SDA,#1
        mov    SCLmask,#1
        shl    SCLmask,SCL
        mov    SDAmask,SCLmask
        shl    SDAmask,#1
        add    r0,#4
        rdlong I2C,r0                           ' get I2C Device to use
        add    r0,#4
        rdlong reg_addr,r0                      ' get address of Registry
        add    r0,#4
        rdlong stack_addr,r0                    ' get address of top of stack
        add    r0,#4
        rdlong entry_addr,r0                    ' get kernel entry address
        add    r0,#4
        rdlong load_addr,r0                     ' get address of Catalina program image in EEPROM (usually $8000)
        add    r0,#4
        rdlong page_addr,r0                     ' get address of Hub RAM page buffer to use (must be long aligned)

        call   #I2C_Setup

        call   #XMM_Activate

        ' retrieve the initial BZ, PC, layout and segment addresses
        ' from the image
        '
        ' Note: the initial BZ and PC in the image will be correct    
        ' once the data and code segments are moved - the BZ value
        ' is correct if we load the data segment to the place the
        ' compiler assumed it would be loaded, and the XMM Kernel
        ' adjusts the PC value to subtract the location of the code
        ' segment, assuming it will end up at location zero in XMM.
        ' If this is not the case, the intial PC value will need to
        ' be adjusted. Note all internal code segment references are
        ' relocated "on the fly" by the XMM Kernel.
        '
        call   #ClearPage                        ' for debug only - make sure page buffer is empty
        mov    addrReg,load_addr                 ' get address of image (in EEPROM)
        add    addrReg,#$10 + INIT_BZ_OFF<<2 + 8 ' point to initial BZ (offset $15C)
        mov    dataPtr,page_addr
        mov    count,#PAGE_SIZE
        call   #I2C_ReadPage
        mov    r0,page_addr
        rdlong Init_BZ,r0
        add    r0,#4                             
        rdlong Init_PC,r0
        add    r0,#4                             
        rdlong layout,r0                        
        add    r0,#4
        rdlong code_addr,r0
        add    r0,#4
        rdlong cnst_addr,r0
        add    r0,#4
        rdlong init_addr,r0
        add    r0,#4
        rdlong data_addr,r0
        add    r0,#4
        rdlong ends_addr,r0
'
' NOTE: the following code implements:
' XMM Layout 2:
'    [CNST,INIT,DATA,CODE] => [CODE] in XMM & [CNST,INIT,DATA] in HUB
' XMM Layout 3:
'    [INIT,DATA,CODE,CNST] => [CODE,CNST] in XMM & [INIT,DATA] in HUB
' XMM Layout 4:
'    [DATA,CODE,CNST,INIT] => [CODE,CNST,INIT] in XMM & [DATA] in HUB
' XMM Layout 5:
'    [CODE,CNST,INIT,DATA] => [CODE,CNST,INIT,DATA] in XMM (stack in HUB)
'

        ' Copy from the start of the code segment to the end 
        ' of all program segments from I2C to XMM
        
        mov    r0,ends_addr
        sub    r0,code_addr
        add    r0,#PAGE_SIZE - 1
        mov    r1,#PAGE_SIZE
        call   #f_d32u
        mov    page_count,r0

        mov    src_addr,load_addr
        add    src_addr,code_addr
        add    src_addr,#$10
        cmp    layout,#5 wz                     ' for layout 5, correct ...
   if_z sub    src_addr,l5size                  ' ... for bytes removed
        mov    dst_addr,#0                      ' code goes at offset 0 in XMM
        
        call   #Copy_I2C_To_XMM

        ' Then copy the first $200 bytes of image from I2C to RAM. Note that
        ' we don't overwrite the first $10 bytes since they're special - i.e.
        ' clock freq etc - so we only copy from byte $10 onwards

        mov    page_count,#$200/PAGE_SIZE       ' ... only $200 bytes

        ' The segments logically go at offset 0 in RAM, but we don't
        ' overwrite the first $10 bytes since they're special - i.e.
        ' clock freq etc - so we only copy from byte $10 onwards

        mov    src_addr,load_addr
        add    src_addr,#$10
        mov    dst_addr,#$10

        call   #Copy_I2C_To_RAM

        ' Now relocate the segments from I2C to RAM (depending on the layout)

        cmp    layout,#5 wz                     ' for layout 5 ...
  if_z  jmp    #reloc_done                      ' ... no segments to relocate
        mov    r0,Init_BZ                       ' for other layouts we copy (up to) Init_BZ
        cmp    layout,#4 wz                     ' for layout 4 ...
  if_z  mov    src_addr,data_addr               ' ... start from data segment
        cmp    layout,#3 wz                     ' for layout 3 ...            
  if_z  mov    src_addr,init_addr               ' ... start from init segment 
        cmp    layout,#2 wz                     ' for layout 2 ...            
  if_z  mov    src_addr,cnst_addr               ' ... start from cnst segment 

        sub    r0,src_addr                      ' calculate size to copy
        mov    dst_addr,src_addr                ' destination addr in RAM is correct ...
        add    src_addr,#$10                    ' ... but we must correct source addr
        add    src_addr,load_addr

        add    r0,#PAGE_SIZE - 1
        mov    r1,#PAGE_SIZE
        call   #f_d32u
        mov    page_count,r0

        call   #Copy_I2C_To_RAM
        
reloc_done

        ' Copy MAX_KERNEL_LONGS*4 bytes from the Kernel entry address  
        ' to straight after Init_BZ (except for layout 5, which copies
        ' the kernel to Just after the segment table) - this is just a 
        ' temporary place to put the kernel code - later, we will 
        ' overwrite ourselves with this code when we restart this cog. 

        mov    r0,kernel_max
        add    r0,#PAGE_SIZE - 1
        mov    r1,#PAGE_SIZE
        call   #f_d32u
        mov    page_count,r0

        mov    src_addr,entry_addr
        cmp    layout,#5 wz                     ' for layout 5 ...
   if_z mov    dst_addr,#$10+INIT_PC_OFF<<2+8+28 ' ... copy after seg table
  if_nz mov    dst_addr,Init_BZ                 ' for other layouts ...
  if_nz add    dst_addr,#$10                    ' ... copy to Init_BZ (+$0010)

        call   #Copy_I2C_To_RAM

        ' Set up the Catalina program base address for use by the Kernel.

        cogid   r0                              ' convert ...
        shl     r0,#2                           ' ... my cog id ...
        add     r0,reg_addr                     ' ... to my registration addr
        rdlong  r2,r0                           ' get my request block addr

        add     r2,#4
        wrlong  stack_addr,r2
        sub     r2,#4
        mov     r1,#$10                         ' initial BA is always $10
        wrlong  r1,r2

        ' Then become the Kernel by restarting the cog with the Kernel code.

        cogid   r0                              ' set the cog id
        cmp     layout,#5 wz                    ' for layout 5 ...
   if_z mov     r1,#$10+INIT_PC_OFF<<2+8+28     ' ... kernel is after seg table 
  if_nz mov     r1,Init_BZ                      ' for other layouts ...
  if_nz add     r1,#$10                         ' ... it is at Init_BZ (+$0010)
        shl     r1,#2
        or      r0,r1                           ' set the code address
        mov     r1,reg_addr
        shl     r1,#16
        or      r0,r1                           ' set the par address

       ' call XMM_TriState to avoid side effects (e.g. resetting 
       ' other CPUs on the TriBladeProp) while restarting

        call #XMM_Tristate
                      
        coginit r0

        ' the following constant is used when reconstructing layout 5:

l5size  long   $8000 - $0200                    ' this many bytes removed

low24   long    $00FFFFFF
        
'-------------------------------- Utility routines -----------------------------


Copy_I2C_To_XMM
        cmp    page_count,#0 wz
 if_z   jmp    #Copy_I2C_To_XMM_ret
:Copy_loop
        mov    addrReg,src_addr
        mov    dataPtr,page_addr
        mov    count,#PAGE_SIZE
        call   #I2C_ReadPage
        mov    XMM_Addr,dst_addr
        mov    Hub_Addr,page_addr
        mov    XMM_Len,#PAGE_SIZE
        call   #XMM_WritePage
        add    src_addr,#PAGE_SIZE
        add    dst_addr,#PAGE_SIZE
        djnz   page_count,#:Copy_loop
Copy_I2C_To_XMM_ret
        ret


Copy_I2C_To_RAM
        cmp    page_count,#0 wz
 if_z   jmp    #Copy_I2C_To_RAM_ret
:Copy_loop
        mov    addrReg,src_addr
        mov    dataPtr,page_addr
        mov    count,#PAGE_SIZE
        call   #I2C_ReadPage
        mov    r0,#PAGE_SIZE/4
        mov    r1,page_addr
:Write_loop
        rdlong r2,r1
        wrlong r2,dst_addr
        add    r1,#4
        add    dst_addr,#4
        djnz   r0,#:Write_loop
        add    src_addr,#PAGE_SIZE
        djnz   page_count,#:Copy_loop
Copy_I2C_To_RAM_ret
        ret


'
'f_d32u - Unsigned 32 bit division
'         Divisor : r1
'         Dividend : r0
'         Result:
'             Quotient in r0
'             Remainder in r1

f_d32u
        mov    ftemp,#32
        mov    ftmp2, #0
:up2
        shl    r0,#1       WC
        rcl    ftmp2,#1    WC
        cmp    r1,ftmp2    WC,WZ
 if_a   jmp    #:down
        sub    ftmp2,r1
        add    r0,#1
:down
        sub    ftemp, #1   WZ
 if_ne  jmp    #:up2
        mov    r1,ftmp2
f_d32u_ret
        ret


'
'f_m32 - multiplication
'        r0 : 1st operand (32 bit)
'        r1 : 2nd operand (32 bit)
'        Result:
'           Product in r0 (<= 32 bit)
'
f_m32
        mov    ftemp,#0
:start
        cmp    r0,#0       WZ
 if_e   jmp    #:down3
        shr    r0,#1       WC
 if_ae  jmp    #:down2
        add    ftemp,r1    WC
:down2
        shl    r1,#1       WC
        jmp    #:start
:down3
        mov    r0,ftemp
f_m32_ret
        ret


ClearPage
        mov    r0,#0
        mov    r1,#PAGE_SIZE
        mov    r2,page_addr
:Clear_loop
        wrbyte r0,r2
        add    r2,#1
        djnz   r1,#:Clear_loop
ClearPage_ret
        ret

'-------------------------------------------------------------------------------
{
#include "debug.inc"
}
'-----------------------------------I2C routines -------------------------------

I2C_Delay
        mov    I2C_Count,cnt
        add    I2C_Count,I2C_Wait
        waitcnt I2C_Count,#0
I2C_Delay_ret
        ret
I2C_Count
        long   $0
I2C_Wait
        long   I2C_DELAY_COUNT+13                ' must be at least 13


I2C_SCL_On
        or     outa,SCLmask                      ' Drive SCL High
        or     dira,SCLmask
I2C_SCL_On_ret
        ret


I2C_SCL_Off
        andn   outa,SCLmask                      ' Drive SCL Low
        or     dira,SCLmask
I2C_SCL_Off_ret
        ret


I2C_SCL_Input
        andn   dira,SCLmask                      ' Set SCL as Input (or to float due to pullup)
I2C_SCL_Input_ret
        ret


I2C_SDA_On
        or     outa,SDAmask                      ' Drive SDA High
        or     dira,SDAmask
I2C_SDA_On_ret
        ret


I2C_SDA_Off
        andn   outa,SDAmask                      ' Drive SDA Low
        or     dira,SDAmask
I2C_SDA_Off_ret
        ret


I2C_SDA_Input
        andn   dira,SDAmask                      ' Set SDA as Input (or to float due to pullup)
I2C_SDA_Input_ret
        ret


I2C_Setup
        call   #I2C_SCL_On                       ' Drive SCL High
        call   #I2C_SDA_Input                    ' Set SDA as Input
        mov    r0,#9                             ' repeat 9 times
:Setup_loop
        call   #I2C_Delay                        '   #### delay ###
        call   #I2C_SCL_Off                      '   Pulse Clock
        call   #I2C_Delay                        '   #### delay ###
        call   #I2C_SCL_On
        mov    r1,ina
        test   r1,SDAmask wz                     '    if ina[SDA]      ' Repeat if SDA not driven high
 if_nz  jmp    #I2C_Setup_ret                    '       quit          '  by the EEPROM
        djnz   r0,#:Setup_loop                   '
I2C_Setup_ret
        ret

I2C_Start
        call   #I2C_SDA_On                       ' Initially drive SDA HIGH
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SCL_On                       ' Initially drive SCL HIGH
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SDA_Off                      ' Now drive SDA LOW
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SCL_Off                      ' Leave SCL LOW
        call   #I2C_Delay                        ' #### delay ###
I2C_Start_ret
        ret


I2C_Stop
        call   #I2C_SDA_Off                      ' Drive SDA LOW
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SCL_On                       ' Drive SCL HIGH
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SDA_On                       ' Then drive SDA HIGH
        call   #I2C_Delay                        ' #### delay ###
'       call   #I2C_SCL_Input                    ' Now let them float (Note - differs from Mike's original driver)
'       call   #I2C_SDA_Input                    ' If pullups present, they'll stay HIGH (Note - differs from Mike's original driver)
I2C_Stop_ret
        ret


I2C_Write
        mov    ackbit,#0                         ' ackbit := 0
        mov    r1,data
        shl    r1,#24                            ' data <<= 24
        mov    r0,#8                             ' repeat 8            ' Output data to SDA
:Write_bit
        rol    r1,#1
        test   r1,#1 wz
 if_z   call   #I2C_SDA_Off
 if_nz  call   #I2C_SDA_On                       '   outa[SDA] := (data <-= 1) & 1
        call   #I2C_SCL_On                       '   Toggle SCL from LOW to HIGH to LOW
        call   #I2C_Delay                        '   #### delay ###
        call   #I2C_SCL_Off                      '
        djnz   r0,#:Write_bit
        call   #I2C_SDA_Input                    ' Set SDA to input for ACK/NAK
        call   #I2C_SCL_On                       ' Sample SDA when SCL is HIGH
        call   #I2C_Delay                        ' #### delay ###
        mov    ackbit,ina
        and    ackbit,SDAmask                    ' ackbit := ina[SDA]
        shr    ackbit,SDA
        call   #I2C_SCL_Off
'       call   #I2C_SDA_Off                      ' Leave SDA driven LOW (Note - differs from Mike's original driver) !!! ### !!! ###
I2C_Write_ret
        ret



I2C_Read
        mov    data,#0                           ' data := 0
        call   #I2C_SDA_Input                    ' Make SDA an Input
        mov    r0,#8                             ' repeat 8 times to recive a byte from SDA
:Read_bit
        call   #I2C_SCL_On                       ' Sample SDA when SCL is HIGH
        call   #I2C_Delay                        '   #### delay ###
        mov    r1,ina
        and    r1,SDAmask
        shr    r1,SDA
        shl    data,#1
        or     data,r1                          '   data := (data << 1) | ina[SDA]
        call   #I2C_SCL_Off                      '   Drive SCL LOW
        call   #I2C_Delay                        '   #### delay ###
        djnz   r0,#:Read_bit
        mov    r1,ackbit wz
 if_z   call   #I2C_SDA_Off
 if_nz  call   #I2C_SDA_On                       ' outa[SDA] := ackbit ' Output ACK/NAK to SDA
        call   #I2C_SCL_On                       ' Toggle SCL from LOW to HIGH to LOW
        call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SCL_Off
'       call   #I2C_SDA_Off                      ' Leave SDA Low (Note - differs from Mike's original driver) !!! ### !!! ###
'       call   #I2C_Delay                        ' #### delay ###
        call   #I2C_SDA_On                       ' Leave SDA High (Note - differs from Mike's original driver) !!! ### !!! ###
I2C_Read_ret
        ret


'
' I2C_ReadPage : Read in a block of i2c data.
'
' Parameters:
'     devSel  : Device select code.
'     addrReg : Device starting address.
'     dataPtr : Data address.
'     count   : Number of bytes.
' Returns:
'     The acknowledge bits if an error occurred.
'
' NOTE : The device select code is modified using the upper 3 bits of the 19 bit addrReg.
'
I2C_ReadPage
        mov    r0,addrReg
        shr    r0,#15
        and    r0,#%1110
        mov    devSel,I2C
        or     devSel,r0                        ' devSel |= addrReg >> 15 & %1110
        call   #I2C_Start                       ' Start(SCL)           ' Select the device & send address
        mov    data,devSel
        or     data,#Xmit
        call   #I2C_Write
        mov    ackbitC,ackbit                   ' ackbit := Write(SCL, devSel | Xmit)
        mov    data,addrReg
        shr    data,#8
        and    data,#$ff
        call   #I2C_Write
        shl    ackbitC,#1
        or     ackbitC,ackbit                   ' ackbit := (ackbit << 1) | Write(SCL, addrReg >> 8 & $FF)
        mov    data,addrReg
        and    data,#$ff
        call   #I2C_Write
        shl    ackbitC,#1
        or     ackbitC,ackbit                   ' ackbit := (ackbit << 1) | Write(SCL, addrReg & $FF)
        call   #I2C_Start                       ' Start(SCL)           ' Reselect the device for reading
        mov    data,devSel
        or     data,#Recv
        call   #I2C_Write
        shl    ackbitC,#1
        or     ackbitC,ackbit                   ' ackbit := (ackbit << 1) | Write(SCL, devSel | Recv)
        mov    r2,count                         ' repeat count - 1 times
        sub    r2,#1 wz
 if_z   jmp    #:ReadPage_last
:ReadPage_loop
        mov    ackbit,#ACK
        call   #I2C_Read
        wrbyte data,dataPtr
        add    dataPtr,#1                       '   byte[dataPtr++] := Read(SCL, ACK)
        djnz   r2,#:ReadPage_loop
:ReadPage_last
        mov    ackbit,#NAK
        call   #I2C_Read
        wrbyte data,dataPtr
        add    dataPtr,#1                       ' byte[dataPtr++] := Read(SCL, NAK)
        call   #I2C_Stop                        ' Stop(SCL)
        mov    ackbit,ackbitC
I2C_ReadPage_ret
        ret                                     ' return ackbit

{
' not required for this loader ...
'

I2C_ReadByte
        mov   count,#1
        call  #I2C_ReadPage
        cmp   ackbit,#0 wz                      ' if ReadPage(SCL, devSel, addrReg, @data, 1)
I2C_ReadByte_ret
        ret                                     '   return -1


I2C_ReadWord
        mov   count,#2
        call  #I2C_ReadPage
        cmp   ackbit,#0 wz                      ' if ReadPage(SCL, devSel, addrReg, @data, 2)
I2C_ReadWord_ret
        ret                                     '   return -1


I2C_ReadLong
        mov   count,#4
        call  #I2C_ReadPage
        cmp   ackbit,#0 wz                      ' if ReadPage(SCL, devSel, addrReg, @data, 4)
I2C_ReadLong_ret
        ret                                     '   return -1
}

{
' not required for this loader ...
'
'
' I2C_WritePage : Write a block of i2c data. NOT TESTED !!!
'
' Parameters:
'     devSel  : Device select code.
'     addrReg : Device starting address.
'     dataPtr : Data address.
'     count   : Number of bytes.
' Returns:
'     The acknowledge bits if an error occurred.
'
' NOTE : The device select code is modified using the upper 3 bits of the 19 bit addrReg.
'
I2C_WritePage
        mov    r0,addrReg
        shr    r0,#15
        and    r0,#%1110
        mov    devSel,I2C
        or     devSel,r0                        ' devSel |= addrReg >> 15 & %1110
        call   #I2C_Start                       ' Start(SCL)           ' Select the device & send address
        mov    data,devSel
        or     data,#Xmit
        call   #I2C_Write
        mov    ackbitC,ackbit                   ' ackbit := Write(SCL, devSel | Xmit)
        mov    data,addrReg
        shr    data,#8
        and    data,#$ff
        call   #I2C_Write
        shl    ackbitC,#1
        or     ackbitC,ackbit                   ' ackbit := (ackbit << 1) | Write(SCL, addrReg >> 8 & $FF)
        mov    data,addrReg
        and    data,#$ff
        call   #I2C_Write
        shl    ackbitC,#1
        or     ackbitC,ackbit                   ' ackbit := (ackbit << 1) | Write(SCL, addrReg & $FF)
        mov    r2,count wz                      ' repeat count times
 if_z   jmp    #:WritePage_done
:WritePage_loop
        and    ackbitC,TOP_BIT wz
        shl    ackbitC,#1
 if_nz  or     ackbitC,TOP_BIT                  ' "Sticky" sign bit
        rdbyte data,dataPtr
        add    dataPtr,#1                       '   Write(SCL, byte[dataPtr++])
        call   #I2C_Write                       ' 
        or     ackbitC,ackbit
        djnz   r2,#:WritePage_loop
:WritePage_done
        call   #I2C_Stop                        ' Stop(SCL)
        mov    ackbit,ackbitC
I2C_WritePage_ret
        ret                                     ' return ackbit



I2C_WriteByte
        mov   count,#1
        call  #I2C_WritePage
        cmp   ackbit,#0 wz                      ' if WritePage(SCL, devSel, addrReg, @data, 1)
I2C_WriteByte_ret
        ret                                     '   return -1


I2C_WriteWord
        mov   count,#2
        call  #I2C_WritePage
        cmp   ackbit,#0 wz                      ' if WritePage(SCL, devSel, addrReg, @data, 2)
I2C_WriteWord_ret
        ret                                     '   return -1


I2C_WriteLong
        mov   count,#4
        call  #I2C_WritePage
        cmp   ackbit,#0 wz                      ' if WritePage(SCL, devSel, addrReg, @data, 4)
I2C_WriteLong_ret
        ret                                     '   return -1
}

'
' Common variables
'
r0            long      $0
r1            long      $0
r2            long      $0
r3            long      $0
r4            long      $0
r5            long      $0
r6            long      $0

page_addr     long      $0
load_addr     long      $0
reg_addr      long      $0
stack_addr    long      $0
entry_addr    long      $0

Init_BZ       long      $0
Init_PC       long      $0

layout        long      $0
code_addr     long      $0
cnst_addr     long      $0
init_addr     long      $0
data_addr     long      $0
ends_addr     long      $0

code_size     long      $0
cnst_size     long      $0
init_size     long      $0
data_size     long      $0

src_addr      long      $0
dst_addr      long      $0
page_count    long      $0
'
' temporary storage used in mul & div calculations
'
ftemp         long      $0
ftmp2         long      $0
ftmp3         long      $0

kernel_max    long      MAX_KERNEL_LONGS*4

'-------------------------------------------------------------------------------
'
' I2C-specific variables
'
SCL           long      $0
SDA           long      $0
I2C           long      $0

SCLmask       long      $0
SDAmask       long      $0

ackbit        long      $0
ackbitC       long      $0                      ' cumulative ackbit
devSel        long      $0
data          long      $0
addrReg       long      $0
dataPtr       long      $0
count         long      $0

'---------------------------------
{

PUB WriteWait(SCL, devSel, addrReg) : ackbit
'' Wait for a previous write to complete.  Device select code is devSel.  Device
'' starting address is addrReg.  The device will not respond if it is busy.
'' The device select code is modified using the upper 3 bits of the 18 bit addrReg.
'' This returns zero if no error occurred or one if the device didn't respond.
   devSel |= addrReg >> 15 & %1110
   Start(SCL)
   ackbit := Write(SCL, devSel | Xmit)
   Stop(SCL)
   return ackbit
}
'
'=============================== XMM SUPPORT CODE ==============================
'
' The folling defines determine which XMM functions are included - comment out
' the appropriate lines to exclude the corresponding XMM function:
'
'#define NEED_XMM_READLONG
'#define NEED_XMM_WRITELONG
'#define NEED_XMM_READPAGE
#define NEED_XMM_WRITEPAGE
#define NEED_XMM_INCADDR        ' required for XMM_ReadPage and XMM_WritePage
#define ACTIVATE_INITS_XMM      ' the HX512 does not allow this in the Kernel
'                                    
#include "XMM.inc"
'
'============================ END OF XMM SUPPORT CODE ==========================
'
              fit       $1f0                    ' max size
'              
{{
                            TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
