{{
'-------------------------------------------------------------------------------
'
'  Catalina_XMM_RamTest - A Memory Test program for XMM RAM.
'  
'  Load the program. It will print "START?" and wait for the "Y" (or "y") key
'  to be pressed. It will then read/write the first few locations of XMM RAM
'  as a trivial test showing that it can read/write XMM RAM at all.
'
'  This should print a few lines like:
'
'     A 000000xx 000000xx DEADBEEF
'
'  Everything is ok if the first two values are the same, and the second value is
'  DEADBEEF, otherwise the program cannot read/write XMM RAM even when no paging
'  is required. The test will continue in any case, but is doomed to fail.
'
'  The trivial test will print AGAIN when complete. Press Y to repeat the trivial
'  test, or N to proceed to a more complex test.
'
'  After the trivial test comes the real test. The program initializes the RAM 
'  and repeatedly reads it back in each of multiple test passes. After each pass
'  it will print "DONE". If any memory cells fail it will print "FAIL" and each
'  XMM memory failure. Note the XMM base address is not included in the address
'  printed, so XMM address 00000000 is always the first XMM location. The base
'  address of the XMM RAM must be set in Catalina_Common.
'
'  After each pass it will print "AGAIN?". If you want to run another pass, press
'  any key. If you press a key in the middle of a pass, the program will print
'  "ABORT", abandon the current pass, and return to the "AGAIN?" prompt.
'
'  Normally, the complex test prints nothing for successful passes. But on each
'  memory failure the program prints:
'  
'      "FAIL XXXXXXXX=YYYYYYYY(ZZZZZZZZ)"
'
'  which means XMM address XXXXXXXX contained the value YYYYYYYY, and not the
'  expected value of ZZZZZZZZ.
'
'  Note that once a memory location fails, it will fail every pass from then on
'  if the memory location has actually been corrupted. If you continue to test,
'  more memory locations may fail. Note that even if the memory is faulty, it may
'  take many passes to fail - it often depends on the size of the test. A test of
'  a few hundred longs may take many dozens of passes to make a location fail,
'  while a large test (more than a few thousand longs) may show a failure on the
'  first pass. Set the number of longs to include in the test using the TEST_SIZE
'  constant.
'
'  A memory test of only a small number of longs may never fail - especially on
'  page-based XMM systems where the problem is with the paging code and the number
'  of longs is small enough that the program never needs to change the XMM page.
'
'  Note that this program tests long writes, long reads and byte writes, and also
'  page reads and writes - but a failure in a page read or write will only be
'  detected during the next test pass - so you should always do at least 2 passes. 
'
'  Normally we would use a PC terminal emulator for I/O (e.g. the Parallax Serial
'  Terminal) so that we don't lose the output as soon as it scrolls off the screen.
'
'   
' Version 2.5 - Add RamBlade support and minor tweaks to XMM code.
' Version 2.8 - Minor changes to the trivial test, and add support for SPI Ram on
'               Morpheus and the C3. 

'-------------------------------------------------------------------------------
}}
CON

TRIVIAL_TEST_SIZE = $10 

#ifdef HYDRA
TEST_SIZE  = $20000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef HYBRID
TEST_SIZE  = $20000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef TRIBLADEPROP
#ifdef CPU_1
TEST_SIZE  = $20000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef CPU_2
TEST_SIZE  = $40000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#else
  ERROR : CPU_1 or CPU_2 must be specified
#endif
#elseifdef RAMBLADE
TEST_SIZE  = $20000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef DRACBLADE
TEST_SIZE  = $20000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef MORPHEUS
#ifdef CPU_1
TEST_SIZE  = $02000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#elseifdef CPU_2
TEST_SIZE  = $A0000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#else
  ERROR : CPU_1 or CPU_2 must be specified
#endif
#elseifdef C3
TEST_SIZE  = $04000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#else
TEST_SIZE  = $A0000 ' number of LONGs to test - must be EVEN (due to address interleaving)
#endif
PAGE_SIZE  = 31     ' number of BYTEs in page for testing page read/write (best if a prime number)

PAGES      = (TEST_SIZE * 4) / PAGE_SIZE ' number of pages to read/write

XMM_BASE_ADDRESS = Common#XMM_BASE_ADDRESS ' Base address of XMM RAM (set in Catalina_Common)

'#define INVERT ' invert test pattern stored in each cell - may help uncover memory problems.

LMM_HMI = Common#LMM_HMI

CON

'
' Set LOAD_RESERVE_COG if we need it
'
#ifdef RESERVE_COG
LOAD_RESERVE_COG  = 1
RESERVE_BYTE_SIZE = RSV#BYTE_SIZE 
#else
LOAD_RESERVE_COG  = 0
RESERVE_BYTE_SIZE = 0
#endif

' HX512 XMM Constants (only used on HX512)

HX_DBUS_BIT_SHIFT    = 16
HX_CTRL_BIT_SHIFT    = 1
HX_STROBE_BIT_SHIFT  = 30

' commands for controller, pre-shifted to left 1-bit, so we don't have to do it
' during runtime, small enough to fit in constants (only used on HX512)

HX_CMD_WRITE_SHIFT   = %00000_00_0
HX_CMD_READ_SHIFT    = %00000_01_0
HX_CMD_LOADLO_SHIFT  = %00000_10_0
HX_CMD_LOADHI_SHIFT  = %00000_11_0

' sram commands (only used on HX512)

HX_CMD_WRITE   = %000000_00
HX_CMD_READ    = %000000_01
HX_CMD_LOADLO  = %000000_10
HX_CMD_LOADHI  = %000000_11

OBJ
   Common : "Catalina_Common"

#ifdef RESERVE_COG
   RSV    : "Catalina_Reserved_Cog"
#endif   

PUB StartReserveCog(Block_Ptr)

#ifdef RESERVE_COG
   RSV.Start(Block_Ptr)
#endif

PUB Run(Xfer_Buffer, Page_Buffer) | Reg, Xfer, Page
'
' Initiate the XMM Tester 
'

  Reg    := Common#REGISTRY
  Xfer   := Xfer_Buffer
  Page   := Page_Buffer

  coginit(cogid, @entry, @Reg)

DAT
        org 0

entry
        mov    r0,par           ' point to parameters
        rdlong reg_addr,r0      ' get registry address
        add    r0,#4
        rdlong xfer_addr,r0     ' get xfer buffer address
        add    r0,#4
        rdlong page_addr,r0     ' get page buffer address

        call   #XMM_Activate    ' Set up the XMM hardware

        call   #delay_2sec
        call   #HMI_KeyClear

trivial_start                 
        call   #HMI_CrLf
        mov    r2,#"S"
        call   #HMI_Char         
        mov    r2,#"T"
        call   #HMI_Char         
        mov    r2,#"A"
        call   #HMI_Char         
        mov    r2,#"R"
        call   #HMI_Char         
        mov    r2,#"T"
        call   #HMI_Char         
        mov    r2,#"?"
        call   #HMI_Char         
        call   #test_for_y
  if_nz jmp    #trivial_start

'{
' trivial test to make sure read/write is working at all (no paging required)
trivial_test
        mov    Count,TrivSize
        mov    Addr,#0
        movs   XMM_Src,#Data
        movd   XMM_Dst,#Data
trivial_loop
        mov    r2,#"A"
        call   #HMI_Char        
        call   #HMI_Space       
        mov    Data,Addr
        mov    XMM_Addr,Addr
        mov    r2,XMM_Addr
        call   #HMI_Hex         
        call   #HMI_Space       
        call   #XMM_WriteLong
        mov    Data,#0
        mov    XMM_Addr,Addr
        call   #XMM_ReadLong
        mov    r2,Data
        call   #HMI_Hex         
        call   #HMI_Space       
        mov    Data,TestVal
        mov    XMM_Addr,Addr
        call   #XMM_WriteLong
        mov    Data,#0
        mov    XMM_Addr,Addr
        call   #XMM_ReadLong
        mov    r2,Data
        call   #HMI_Hex         
        call   #HMI_Space       
        call   #HMI_CrLf
        add    Addr,#4
        djnz   Count,#trivial_loop

trivial_again
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"G"
        call   #HMI_Char                    
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"I"
        call   #HMI_Char                    
        mov    r2,#"N"
        call   #HMI_Char                    
        mov    r2,#"?"
        call   #HMI_Char                    
        call   #test_for_y
   if_z jmp    #trivial_test


' code to initializa all RAM longs with test data
        mov    Count,TestSize
        mov    Addr,#0
        movs   XMM_Src,#Data
        movd   XMM_Dst,#Data
init_loop
        call   #Calc_XMM_Addr
#ifdef INVERT        
        neg    Data,Count
#else
        mov    Data,Count
#endif        
        call   #XMM_WriteLong
        add    Addr,#4
        djnz   Count,#init_loop
        
{
' enable this code to print the value assigned to each memory cell AFTER initialization
        mov    Addr,TestSize
        shl    Addr,#2
        mov    Count,TestSize
        movs   XMM_Src,#Data
        movd   XMM_Dst,#Data

init_loop_2
        sub    Addr,#4
#ifdef INVERT        
        neg    Data,Count
#else
        mov    Data,Count
#endif
        mov    XMM_Addr,Addr        
        call   #XMM_ReadLong
        mov    r2,#"C"           
        call   #HMI_Char        
        call   #HMI_Space       
        mov    r2,Addr
        call   #HMI_Hex         
        call   #HMI_Space       
        mov    r2,Data
        call   #HMI_Hex         
        call   #HMI_CrLf        
        djnz   Count,#init_loop_2
        call   #press_a_key        
}

'
' test_loop - the main test loop - an infinite loop
'
test_loop

{
' enable this code to read longs more intensively during each test pass (may help uncover memory problems)
        mov    Count,TestSize
        mov    Addr,#0
        movd   XMM_Dst,#Data
stress_test
        call   #Calc_XMM_Addr
        call   #XMM_ReadLong
        add    Addr,#4
        djnz   Count, #stress_test
}

        mov    Count,TestSize
        mov    Addr,#0
        movd   XMM_Dst,#Data
'
' test_read - read and compare all values
'
test_read
        call   #Calc_XMM_Addr
        call   #XMM_ReadLong
#ifdef INVERT        
        neg    Value,Count
#else
        mov    Value,Count
#endif        
        cmp    Data,Value wz
 if_nz  jmp    #compare_fail        

compare_ok
{
' enable this code to print the actual value read on each successful comparison       
        mov    r2,#"P"
        call   #HMI_Char                      
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"S"
        call   #HMI_Char                    
        mov    r2,#"S"
        call   #HMI_Char                    
        jmp    #compare_print
}        
{
' enable this code to just print a '.' for each successful comparison       
        mov    r2,#"."               
        call   #HMI_Char            
}
' default is to not print anything on each successful comparison        
        jmp    #test_next

compare_fail
        mov    r2,#"F"
        call   #HMI_Char                    
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"I"
        call   #HMI_Char                    
        mov    r2,#"L"
        call   #HMI_Char                    
compare_print        
        call   #HMI_Space                   
        mov    r2,Addr
        call   #HMI_Hex                     
        mov    r2,#"="
        call   #HMI_Char                    
        mov    r2,Data
        call   #HMI_Hex                     
        mov    r2,#"("
        call   #HMI_Char                    
        mov    r2,Value
        call   #HMI_Hex                     
        mov    r2,#")"
        call   #HMI_Char                    
        call   #HMI_CrLf                    

test_next
        add    Addr,#4
        call   #HMI_KeyReady
        tjnz   r0,#test_abort
        djnz   Count,#test_read

test_pass
        mov    r4,page_count
        mov    r5,#0        

page_loop
        mov    XMM_Addr,r5
        mov    Hub_Addr,page_addr
        mov    XMM_Len,max_page 
        call   #XMM_ReadPage

        mov    r2,r5
        mov    r3,max_page
        movs   XMM_Src,#r0
        mov    r0,#0
zero_page               
        mov    XMM_Addr,r2
        mov    XMM_Len,#1
        call   #XMM_WriteMult
        add    r2,#1
        djnz   r3,#zero_page
        
        mov    XMM_Addr,r5
        mov    Hub_Addr,page_addr
        mov    XMM_Len,max_page 
        call   #XMM_WritePage
        add    r5,max_page
        djnz   r4,#page_loop

        mov    r2,#"D"
        call   #HMI_Char                    
        mov    r2,#"O"
        call   #HMI_Char                    
        mov    r2,#"N"
        call   #HMI_Char                    
        mov    r2,#"E"
        call   #HMI_Char                    
        call   #HMI_CrLf

        jmp    #test_again
        
        
test_abort        
'{        
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"B"
        call   #HMI_Char                    
        mov    r2,#"O"
        call   #HMI_Char                    
        mov    r2,#"R"
        call   #HMI_Char                    
        mov    r2,#"T"
        call   #HMI_Char                    
        call   #HMI_CrLf                    
'}
test_again
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"G"
        call   #HMI_Char                    
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#"I"
        call   #HMI_Char                    
        mov    r2,#"N"
        call   #HMI_Char                    
        mov    r2,#"?"
        call   #HMI_Char                    
        call   #HMI_Space                    
        call   #HMI_KeyGet
        mov    r2,r0                        
        call   #HMI_Char                    
        call   #HMI_CrLf                    

        jmp    #test_loop

press_a_key
        mov    r2,#"P"
        call   #HMI_Char                    
        mov    r2,#"R"
        call   #HMI_Char                    
        mov    r2,#"E"
        call   #HMI_Char                    
        mov    r2,#"S"
        call   #HMI_Char                    
        mov    r2,#"S"
        call   #HMI_Char                    
        mov    r2,#" "
        call   #HMI_Char                    
        mov    r2,#"A"
        call   #HMI_Char                    
        mov    r2,#" "
        call   #HMI_Char                    
        mov    r2,#"K"
        call   #HMI_Char                    
        mov    r2,#"E"
        call   #HMI_Char                    
        mov    r2,#"Y"
        call   #HMI_Char                    
        call   #HMI_CrLf                    
        call   #HMI_KeyGet
press_a_key_ret
        ret
'
' test for "Y" or "y" pressed - return Z set if so
'        
test_for_y        
        call   #HMI_KeyGet
        mov    r2,r0
        mov    r3,r0                        
        call   #HMI_Char                    
        call   #HMI_CrLf
        cmp    r3, #"Y" wz
  if_nz cmp    r3, #"y" wz
test_for_y_ret
        ret
'        
'-------------------------------- Test variables -------------------------------

TrivSize long  TRIVIAL_TEST_SIZE                ' size of trivial test
TestSize long  TEST_SIZE                        ' number of XMM longs to test - must be even!!!         

Data     long  $0000_0000                       ' actual value read/written
Value    long  $0000_0000                       ' value expected

Count    long  0
Addr     long  0        

Delay1   long  80_000_000
TestVal  long  $DEADBEEF  
        
'-------------------------------- Utility routines -----------------------------
'
' Calc_XMM_Addr - calculate interleaving addresses to force many XMM memory page
' changes - i.e. we end up with an interleaved memory usage pattern like the
' following:
'
'   (for 10 cells): 0,9,2,7,4,5,6,3,8
'
Calc_XMM_Addr
        test Addr,#%0000_00100 wz  ' odd address?
 if_z   jmp #Calc_Even
        mov  r0,TestSize
        shl  r0,#2
        sub  r0,Addr
        mov  XMM_Addr,r0
        jmp  #Calc_Print
Calc_Even
        mov  XMM_Addr,Addr
Calc_Print
{
' enable this code to print the actual interleaved XMM address used for each nominal address                
        mov    r2,XMM_Addr
        call   #HMI_Hex                    
        call   #HMI_Space                  
}        
Calc_XMM_Addr_ret
        ret            

' delay one second
delay_1sec        
        mov    r1,cnt           
        add    r1,Delay1
        waitcnt r1,#0
delay_1sec_ret

' delay two seconds
delay_2sec
        call   #delay_1sec        
        call   #delay_1sec        
delay_2sec_ret
        ret
        
'-------------------------------- HMI Routines -------------------------------

' HMI_Command - send a request to HMI Plugin for execution
' On Entry
'          r1 = command
'          r2 = data
'
HMI_Command

#ifdef TRIBLADEPROP
        call    #XMM_Tristate   ' disable XMM while using HMI
#endif

        shl     r1,#24          ' construct ...  
        and     r2,Low24        ' ... 
        or      r2,r1           ' ... request

        mov     r0,#LMM_HMI     ' plugin type we want 

        mov     ftemp,reg_addr  ' point to registry
        mov     ftmp2,#0        ' start at cog 0
send1
        cmp     ftmp2,#8 wc,wz  ' run out of plugins?
 if_ae  jmp     #sendErr        ' yes - no such plugin
        rdlong  ftmp3,ftemp     ' no - check next plugin type
        shr     ftmp3,#24       ' is it ...
        cmp     ftmp3,r0 wz     ' ... the type what we wanted?
 if_z   jmp     #send2          ' yes - use this plugin
        add     ftmp2,#1        ' no ...
        add     ftemp,#4        ' ... check ...
        jmp     #send1          ' ... next cog
send2
        mov     r0,ftmp2        ' use the cog where we found the plugin
        shl     r0,#2           ' multiply plugin (cog) id by 4 ...
        add     r0,reg_addr     ' add registry base to get registry entry
        rdlong  r0,r0           ' get request block from registry
        test    r0,top8 wz      ' plugin registered?
 if_z   jmp     #sendErr        ' no - return error
        and     r0,low24        ' yes - write request ...
        wrlong  r2,r0           ' ... to request block
loop2   rdlong  r2,r0   wz      ' wait till ...
 if_nz  jmp     #loop2          ' ... request completed
        add     r0,#4           ' get results ...
        rdlong  r0,r0           ' ... from request block
        jmp     #sendDone
sendErr
        neg     r0,#1           ' return -1 on any error
sendDone

#ifdef TRIBLADEPROP
        call    #XMM_Activate   ' enable XMM again
#endif

        jmp     #HMI_Command_ret
HMI_Command_ret
HMI_XferCmd_ret
        ret
        
top8    long    $FF000000
low24   long    $00FFFFFF

ftemp   long    $0
ftmp2   long    $0
ftmp3   long    $0

' HMI_XferCmd - send a short request to HMI Plugin for execution
'               using an xfer block to hold the data
' On Entry
'          r1 = command
'          r2 = data
'
HMI_XferCmd
        wrlong r2,xfer_addr     ' put data in xfer block
        mov    r2,xfer_addr     ' use xfer block as data
        or     r2,cursor
        jmp    #HMI_Command     ' then same as short command
'
' HMI_KeyReady - see if key (return in r0, or 0 if not ready)
'
HMI_KeyReady
        mov    r1,#5                         
        call   #HMI_XferCmd                  
HMI_KeyReady_ret
        ret
'
' HMI_KeyClear - see Clear Keyboard
'
HMI_KeyClear
        mov    r1,#6                         
        call   #HMI_XferCmd                  
HMI_KeyClear_ret
        ret
'
' HMI_KeyGet - get a key (return in r0).
' This used to call the HMI function, but that function has now been removed
' due to lack of space in the HMI plugins.
'
HMI_KeyGet
        call   #HMI_KeyReady
        tjz    r0,#HMI_KeyGet
        mov    r1,#2                         
        call   #HMI_XferCmd                  
HMI_KeyGet_ret
        ret
        
HMI_Hex
        mov     number,r2      
        mov     digits,#8       ' print 8 digits
:t_hex1
        rol     number,#4       ' convert 4 bits ...
        mov     r2,number       ' ... to '0' .. '9'
        and     r2,#$f          ' ... or ...
        cmp     r2,#10 wc,wz    ' ... 'A' .. 'F' ...
 if_ae  add     r2,#($41-$30-10)' ... depending ...
        add     r2,#$30         ' ... on the digit value
        call    #HMI_Char       ' write char to screen
        djnz    digits,#:t_hex1 ' continue with next digit
HMI_Hex_ret
        ret
digits long 0
number long 0
'
'
' HMI_Char - send a char (char in r2)
'
HMI_Char
        mov    r1,#22
        or     r2,cursor                        
        call   #HMI_Command                  
HMI_Char_ret
        ret
'
' HMI_Space - send a space
'
HMI_Space
        mov    r2,#32                        
        call   #HMI_Char                     
HMI_Space_ret
        ret
'
' HMI_CrLf - send a CRLF
'
HMI_CrLf
        mov    r2,#13                        
        call   #HMI_Char                     
        mov    r2,#10                        
        call   #HMI_Char                     
HMI_CrLf_ret
        ret
'
'------------------------------- Common Variables ------------------------------
'
r0            long      $0
r1            long      $0
r2            long      $0
r3            long      $0
r4            long      $0
r5            long      $0

reg_addr      long      $0
xfer_addr     long      $0
page_addr     long      $0
page_count    long      PAGES
max_page      long      PAGE_SIZE

src_addr      long      $0
dst_addr      long      $0
end_addr      long      $0

cursor        long      1<<23   ' use cursor 1 (srcolls and has visible cursor)
'
'=============================== XMM SUPPORT CODE ==============================
'
' The folling defines determine which XMM functions are included - comment out
' the appropriate lines to exclude the corresponding XMM function:
'
#define NEED_XMM_READLONG
#define NEED_XMM_WRITELONG
#define NEED_XMM_READPAGE
#define NEED_XMM_WRITEPAGE
#define NEED_XMM_INCADDR        ' required for XMM_ReadPage and XMM_WritePage
#define ACTIVATE_INITS_XMM      ' the HX512 does not allow this in the Kernel
'                                    
#include "XMM.inc"
'
'============================ END OF XMM SUPPORT CODE ==========================
'
              fit       $1f0                    ' max size
'
{{
                            TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
