{{
'-------------------------------------------------------------------------------
'
' XMM_blackcat - an XMM target for Catalina Programs using the BlackCat debugger.
'
' This target uses the XMM Kernel, and must be used with the
' XMM Catalina memory layouts (-x2 to -x5).
'
' This target can be used for programs to be loaded from EEPROM,
' or from SD Card (using the Generic SD Loader).
'
' The following symbols affect the kernel and plugins - to
' define them, use the -D command line option (e.g. -DVGA)
' except for the library symbols, which are defined when the
' appropriate library is included (e.g. -lma, -lmb)
'
'    EEPROM         - Use the EEPROM Loader
'    SDCARD         - Use the SDCARD Loader
'    NO_FP          - do not load any FP plugins
'    NO_FLOAT       - do not load any FP Plugins (new in 2.8)
'    NO_HMI         - do not load any HMI plugin
'    NO_MOUSE       - do not load/start mouse driver
'    NO_KEYBOARD    - do not load/start keyboard driver
'    NO_SCREEN      - do not load/start screen driver
'    HIRES_VGA      - load a HiRes VGA driver
'    LORES_VGA      - load a LoRes VGA driver
'    VGA            -   "  "   "    "    "
'    HIRES_TV       - load a HiRes TV driver 
'    LORES_TV       - load a LoRes TV driver 
'    TV             -   "  "   "    "    "    
'    NO_INTERLACE   - use non-interlace mode (TV drivers only)
'    NTSC           - use NTSC mode (TV drivers only)    
'    PC             - use a PC HMI plugin (screen/kbd only)
'    PROPTERMINAL   - use a PropTerminal HMI plugin
'    CLOCK          - load a Real-Time Clock plugin
'    libma          - generated automatically by -lma option
'    libmb          - generated automatically by -lmb option
'    libcx          - generated automatically by -lcx option
'    PROXY_KEYBOARD - use a proxy for the keyboard driver
'    PROXY_MOUSE    - use a proxy for the mouse driver
'    PROXY_SCREEN   - use a proxy for the screen driver
'    PROXY_SD       - use a proxy for the SD plugin
'
'    PROXY_HMI      - (internal) load the proxy HMI plugin
'    PROXY_SIO      - (internal) load the SIO plugin for use by the proxy plugin
'                 
'    RESERVE_COG    - Load the reserve cog
'                 
' Version 1.0 - initial version by Ross Higson
' Version 2.8 - Added support for reserve cog.
'               Simplified object selection logic.
'               Added NO_FLOAT (equivalent to NO_FP)
'
'-------------------------------------------------------------------------------
}}
CON
'
' Set these to suit the platform by modifying "Catalina_Common"
' Also, that is now where the PIN definitions for the platform
' are defined.
'
_clkmode  = Common#CLOCKMODE
_xinfreq  = Common#XTALFREQ
_stack    = Common#STACKSIZE
'
' Set up Proxy and HMI symbols and constants:
'
#include "Constants.inc"
'
OBJ
'
' The following objects MUST be loaded ...
'
  Common   : "Catalina_Common"                          ' Common Definitions
'
' Include the Loader selection logic and load the appropriate objects:
'
#include "Loader.inc"
'
' The following object MUST be loaded ...
'
  BlackCat : "BlackCat_DebugCog"                        ' BlackCat Debugger Cog (for comms with Host)
'
' Include the Kernel selection logic and load the appropriate objects:
'
#include "Kernel.inc"
'
' Include the HMI selection logic and load the appropriate objects:
'
#include "HMI.inc"
'
' Include the Floating Point selection logic and load the appropriate objects:
'
#include "FP.inc"
'
' Include selection logic for other plugins and load the appropriate objects:
'
#include "Other.inc"
'
PUB Start : ok | cog, ALLOC, HMI_DATA, CGI_DATA, RSV_DATA, PAGE, IO_BLOCK, BC_BLOCK, Proxy_Lock, Server_CPU, s

  Common.InitializeRegistry

#ifdef RESERVE_COG
  ' Point to start of File buffers (we must allocate data blocks down from here)
  ALLOC := Common#FLIST_BUFF
#else  
  ' Point to start of Registry (we must allocate data blocks down from here)
  ALLOC := Common#REQUESTS
#endif

  ok         := TRUE
  Proxy_Lock := -1

#ifdef RESERVE_COG
  ' allocate Hub RAM for reserved cog
  ' (the reserve cog is always allocated space and started first)
  RSV_DATA := ALLOC - RSV#BYTE_SIZE
  ALLOC    := RSV_DATA
  if ok
    ' start the reserved cog
    ok := RSV.Start(RSV_DATA)
#endif

#ifdef NO_ARGS
  CogStore.Stop
#else
  s := CogStore.Size
  if (s => 0)
    ALLOC := ALLOC - s*4
  CogStore.Setup(ALLOC)
  CogStore.Stop 
#endif

#ifdef PROXY_SIO
  ' Figure out the server CPU we should use as proxy
  ' (works on Morpheus & TriBladeProp, may not work on others):
#ifdef CPU_1
  Server_CPU  := 2
#elseifdef CPU_2
  Server_CPU  := 1
#elseifdef CPU_3
  Server_CPU  := 2
#endif
  ' allocate a lock to serialize proxy access
  Proxy_Lock := locknew
  ' allocate a data block for SIO
  IO_BLOCK := ALLOC - SIO#BLOCK_SIZE
  ALLOC    := IO_BLOCK
  ' start the SIO plugin
  if ok
    ok := SIO.Start(Common#REGISTRY, IO_BLOCK, Common#RX_PIN, Common#TX_PIN, MODE, BAUD, TRUE)
#ifdef NO_HMI
#else    
  ' set up the proxy parameters
  HMI.ProxySetup(IO_BLOCK, Proxy_Lock, Server_CPU)
#endif  
#endif

#ifdef START_GRAPHICS
  CGI_DATA := ALLOC - CGI#BYTE_SIZE
  CGI_DATA := 128 * (CGI_DATA / 128) ' must align to a 128 byte boundary
  ALLOC    := CGI_DATA
  ' start the selected CGI Plugin (and graphics drivers)
  if ok
    ok := CGI.Start(Common#REGISTRY, CGI_DATA)
#endif

#ifdef NO_HMI
#else
  ' allocate data block for HMI
  HMI_DATA := ALLOC - 4 * HMI#DATASIZE
  ALLOC    := HMI_DATA
  ' start the selected HMI Plugin (and keyboard, mouse & screen drivers)
  if ok
    ok := HMI.Start(Common#REGISTRY, HMI_DATA, START_KBD, START_MOUSE, START_SCREEN)
#endif

  ' start additional plugins here

#ifdef NO_FP
#elseifdef NO_FLOAT
#else
  if ok
    ok := Float_A.Start(Common#REGISTRY)
#ifdef libmb
  if ok
    ok := Float_B.Start(Common#REGISTRY)
#endif
#endif

#ifdef CLOCK
  if ok
    ' start the real-time clock - the second parameter controls the debug
    ' LED (on the Hydra) : 0 => LED off, 1 => LED toggles once per second
    ok := RTC.Start(Common#REGISTRY, 0)
#endif

#ifdef NEED_SD
  if ok
    ' start the SD Card Support plugin
#ifdef PROXY_SD
    ok := SD.Start(Common#REGISTRY, Common#PROXY_XFER, IO_BLOCK, Proxy_Lock, Server_CPU)
#else  
    ok := SD.Start(Common#REGISTRY)
#endif    
#endif

  ' Start the Black Cat Debugger Cog
  if ok
    BC_BLOCK := ALLOC - Common#BLACKCAT_SIZE 
    ALLOC    := BC_BLOCK
    ok := BlackCat.start(Common#BLACKCAT_RXPIN, Common#BLACKCAT_TXPIN, Common#BLACKCAT_MODE, Common#BLACKCAT_BAUD, BC_BLOCK)  

  PAGE  := ALLOC - Loader#PAGE_SIZE
  ALLOC := PAGE

#ifdef EEPROM
  if ok
    ' Kernel and all plugins now loaded - note that for XMM programs
    ' we don't either load or start the Kernel directly, instead we load
    ' and start the XMM Loader, which will replace itself with the XMM
    ' Kernel code when it has finished loading the Catalina program.
    Loader.Run(Common#REGISTRY, ALLOC, Kernel.Entry_Addr, $8000, PAGE)
#else
  if ok
    ' Kernel and all plugins now loaded - note that for XMM programs
    ' we don't either load or start the Kernel directly, instead we load
    ' and start the XMM Loader, which will replace itself with the XMM
    ' Kernel code when it has finished loading the Catalina program.
    Loader.Run(Common#REGISTRY, ALLOC, Kernel.Entry_Addr, PAGE)
#endif

{{

+------------------------------------------------------------------------------------------------------------------------------+
?                                                   TERMS OF USE: MIT License                                                  ?                                                            
+------------------------------------------------------------------------------------------------------------------------------?
?Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    ? 
?files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    ?
?modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software?
?is furnished to do so, subject to the following conditions:                                                                   ?
?                                                                                                                              ?
?The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.?
?                                                                                                                              ?
?THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          ?
?WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         ?
?COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   ?
?ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         ?
+------------------------------------------------------------------------------------------------------------------------------+
}}  
